/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Authors: Artem Bityutskiy, Jarkko Lavinen
 *
 * WARNING: this test program may kill your flash and your device. Do not
 * use it. Authors are not responsible for any damage caused by this
 * program.
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kobject.h>
#include <linux/device.h>
#include <linux/err.h>
#include <linux/mtd/mtd.h>
#include <linux/sched.h>

#define PRINT_PREF KERN_CRIT "EB torture: "

/* Uncomment this if you have old test
#define writesize oobblock */

static int eb = 20;
module_param(eb, int, S_IRUGO);
MODULE_PARM_DESC(eb, "eraseblock number within the selected MTD device");

static int ebcnt = 32;
module_param(ebcnt, int, S_IRUGO);
MODULE_PARM_DESC(ebcnt, "number of consequtive eraseblocks to torture");

static int dev = 0;
module_param(dev, int, S_IRUGO);
MODULE_PARM_DESC(dev, "MTD device number to use");

static int gran = 1000;
module_param(gran, int, S_IRUGO);
MODULE_PARM_DESC(gran, "how often the status information should be printed");

static int check = 1;
module_param(check, int, S_IRUGO);
MODULE_PARM_DESC(check, "if the written data should be checked");

static unsigned int cycles_count = 0;
module_param(cycles_count, uint, S_IRUGO);
MODULE_PARM_DESC(cycles_count, "how many erase cycles to do (infinite by default)");

static struct mtd_info *mtd;
static unsigned char *patt1, *patt2, *check_buf;
static unsigned int erase_cycles;

static int ebtest(int ebnum);

static int __init tort_init(void)
{
	int err, i, size, infinite = !cycles_count;

	printk(PRINT_PREF "Torture %d eraseblocks (%d-%d) of mtd%d\n",
	       ebcnt, eb, eb + ebcnt, dev);
	printk(PRINT_PREF "Write verify %s\n", check ? "enabled" : "disabled");

	mtd = get_mtd_device(NULL, dev);
	if (IS_ERR(mtd)) {
		err = PTR_ERR(mtd);
		printk(PRINT_PREF "error: Cannot get MTD device\n");
		return err;
	}

	err = -ENOMEM;
	patt1 = kmalloc(mtd->erasesize, GFP_KERNEL);
	if (!patt1) {
		printk(PRINT_PREF "error: cannot allocate momory\n");
		goto out_mtd;
	}

	patt2 = kmalloc(mtd->erasesize, GFP_KERNEL);
	if (!patt2) {
		printk(PRINT_PREF "error: cannot allocate momory\n");
		goto out_patt1;
	}

	check_buf = kmalloc(mtd->erasesize, GFP_KERNEL);
	if (!check_buf) {
		printk(PRINT_PREF "error: cannot allocate momory\n");
		goto out_patt2;
	}

	if (mtd->writesize > 1)
		size = mtd->writesize;
	else
		size = mtd->erasesize;

	for (i = 0; i < mtd->erasesize / size; i++) {
		if (i & 1)
			memset(patt1 + i * size, 0x55, size);
		else
			memset(patt1 + i * size, 0xAA, size);
	}
	for (i = 0; i < mtd->erasesize / size; i++) {
		if (!(i & 1))
			memset(patt2 + i * size, 0x55, size);
		else
			memset(patt2 + i * size, 0xAA, size);
	}

	while (1) {
		int i;

		for(i = eb; i < eb + ebcnt; i++) {
			err = ebtest(i);
			if (err)
				break;
		}

		erase_cycles += 1;

		if (unlikely(erase_cycles % gran == 0))
			printk(KERN_CRIT "%08u erase cycles done\n", erase_cycles);

		if (!infinite && --cycles_count == 0)
			break;

		cond_resched();
	}

	printk(PRINT_PREF "torture finished after %u erase cycles\n",
	       erase_cycles);
	kfree(check_buf);
out_patt2:
	kfree(patt2);
out_patt1:
	kfree(patt1);
out_mtd:
	put_mtd_device(mtd);
	if (err)
		printk(PRINT_PREF "error %d occured during torturing\n", err);
	return err;
}
module_init(tort_init);

static void __exit tort_exit(void)
{
	return;
}
module_exit(tort_exit);

static void report_corrupt(int ebi, unsigned char *buf,
			   unsigned char *check_buf);

/*
 * This function writes erases the eraseblock, checks that erase succeeded,
 * writes pattern there, and checks that write succeeded.
 */
static int ebtest(int ebnum)
{
	struct erase_info ei;
	loff_t addr = ebnum * mtd->erasesize;
	size_t read, written;
 	int err;
	unsigned char *buf;

	memset(&ei, 0, sizeof(struct erase_info));
	ei.mtd  = mtd;
	ei.addr = addr;
	ei.len  = mtd->erasesize;

	err = mtd->erase(mtd, &ei);
	if (unlikely(err)) {
		printk(PRINT_PREF "error %d while erasing EB %d\n", err, ebnum);
		return err;
	}

	if (unlikely(ei.state == MTD_ERASE_FAILED)) {
		printk(PRINT_PREF "some erase error occurred at EB %d\n", ebnum);
		return -EIO;
	}

	if (check) {
		/* Make sure we have only 0xFF bytes */
		memset(check_buf, 0, mtd->erasesize);
		read = 0;
		err = mtd->read(mtd, addr, mtd->erasesize, &read, check_buf);
		if (unlikely(err == -EUCLEAN))
			printk(PRINT_PREF "single bit flip occured at EB %d "
			       "after erase, read %zd. Strange!\n",
			       read, ebnum);
		else if (unlikely(err)) {
			printk(PRINT_PREF "error %d while reading EB %d after"
			       " erase, read %zd\n", err, ebnum, read);
			return err;
		} else if (unlikely(read != mtd->erasesize)) {
			printk(PRINT_PREF "failed to read %d bytes after erase,"
			       " read only %zd\n", mtd->erasesize, read);
			return err;
		}
	}

	if ((eb + erase_cycles) & 1)
		buf = patt1;
	else
		buf = patt2;

	written = 0;
	err = mtd->write(mtd, addr, mtd->erasesize, &written, buf);
	if (unlikely(err)) {
		printk(PRINT_PREF "error %d while writing EB %d, written %zd\n",
		       err, ebnum, written);
		return err;
	}

	if (check) {
		memset(check_buf, 0, mtd->erasesize);
		read = 0;
		err = mtd->read(mtd, addr, mtd->erasesize, &read, check_buf);
		if (unlikely(err == -EUCLEAN))
			printk(PRINT_PREF "single bit flip occured at EB %d, "
			       "read %d\n", ebnum, read);
		else if (unlikely(err)) {
			printk(PRINT_PREF "error %d while reading EB %d, "
			       "read %zd\n", err, ebnum, read);
			return err;
		} else if (unlikely(read != mtd->erasesize)) {
			printk(PRINT_PREF "failed to read %d bytes, read "
			       "only %zd\n", mtd->erasesize, read);
			return err;
		}

		if (unlikely(memcmp(buf, check_buf, mtd->erasesize))) {
			report_corrupt(ebnum, buf, check_buf);
			return -EIO;
		}
	}

	return 0;
}

static int countdiffs(unsigned char *buf, unsigned char *check_buf,
		      unsigned offset, unsigned len, unsigned *bytesp,
		      unsigned *bitsp);
static void print_bufs(void *bufread, void *bufwritten, int start, int len);

/*
 * Report the detailed information about how the read EB differs from what was
 * written.
 */
static void report_corrupt(int ebi, unsigned char *buf,
			   unsigned char *check_buf)
{
	int i;
	int bytes, bits, pages, first;
	int offset, len;

	bytes = bits = pages = 0;
	for(i = 0; i < mtd->erasesize; i += mtd->writesize)
		if (countdiffs(buf, check_buf, i, mtd->writesize, &bytes,
			       &bits) >= 0)
			pages ++;

	printk(PRINT_PREF "Read incorrect data back from EB %d.\n", ebi);
	printk(PRINT_PREF "Verify fails on %d pages, %d bytes/%d bits\n",
	       pages, bytes, bits);

	for (i = 0; i < mtd->erasesize; i += mtd->writesize) {
		bytes = bits = 0;
		first = countdiffs(buf, check_buf, i, mtd->writesize, &bytes,
				   &bits);
		if (first < 0)
			continue;

		printk(PRINT_PREF "Page %d has %d bytes/%d bits failing verify,"
		       " starting at offset 0x%x\n",
		       i / mtd->writesize, bytes, bits, first);

		offset = first & ~0x7;
		len    = ((first + bytes) | 0x7) + 1 - offset;

		print_bufs(check_buf, buf, offset, len);
	}
}

static void print_bufs(void *bufread, void *bufwritten, int start, int len)
{
	int i = 0, j1, j2;
	char *p1 = bufread, *p2 = bufwritten;
	char *diff;
	printk("Offset           Read                   Written\n");
	while(i < len) {
		printk("0x%03x: ", start + i);
		diff = "   ";
		for(j1 = 0; j1 < 8 && i + j1 < len; j1++) {
			printk(" %02x", p1[start + i + j1]);
			if (p1[start + i + j1] != p2[start + i + j1])
				diff = "***";
		}

		while(j1 < 8) {
			printk("   ");
			j1++;
		}

		printk("  %s  ", diff);

		for(j2 = 0; j2 < 8 && i + j2 < len; j2++)
			printk(" %02x", p2[start + i + j2]);
		printk("\n");
		i += 8;
	}
}

/*
 * Count the number of differing bytes and bits and return the first differing
 * offset.
 */
static int countdiffs(unsigned char *buf, unsigned char *check_buf,
		      unsigned offset, unsigned len, unsigned *bytesp,
		      unsigned *bitsp)
{
	unsigned i, bit;
	int first = -1;

	for (i = offset; i < offset + len; i++)
		if (buf[i] != check_buf[i]) {
			first = i;
			break;
		}

	while(i < offset + len) {
		if (buf[i] != check_buf[i]) {
			(*bytesp) ++;
			bit = 1;
			while(bit < 256) {
				if ((buf[i] & bit) != (check_buf[i] & bit))
					(*bitsp) ++;
				bit <<= 1;
			}
		}
		i++;
	}

	return first;
}


MODULE_DESCRIPTION("Eraseblock torturing module");
MODULE_AUTHOR("Artem Bityutskiy, Jarkko Lavinen");
MODULE_LICENSE("GPL");
